// FTS_MFX.cpp : Defines the basic functionality of the "Fractal Tune Smithy".
//

#include "stdafx.h"
#include <InitGuid.h>	// cause MidiFilter.h to define GUIDs in this module (vs. merely declare)
#include <MidiFilter.h>
#include "FTS_MFX.h"
#include "FTS_MFXPropPage.h"
#include "externs.h"
#include <time.h>
char do_all_sound_off;
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
long g_lServerLocks = 0;
long g_lComponents = 0;
PROPERTIES prop;
char log_header_up_to_date;
char started_log;
char do_pb_test;
IMfxInputPulse*	g_pInputPulse;
char chann_change_since_start_play;
char szLogPFile[MAX_PATH];
#ifdef _DEBUG
 #ifdef DO_REDO_LOG
  #define REDO_LOG\
  if(fplog)\
  {\ 
   fclose(fplog);\
   fplog=fopen(szLogPFile,"at");\
  }
 #else
  #define REDO_LOG
 #endif
#else
 #define REDO_LOG
#endif

/////////////////////////////////////////////////////////////////////////////
//
// CFTS_MFX
//
/////////////////////////////////////////////////////////////////////////////

// Constructors/Deconstructors

CFTS_MFX::CFTS_MFX() :
	m_cRef( 1 ),
	m_bDirty( FALSE )
{ 
	::InterlockedIncrement( &g_lComponents );
	::InitializeCriticalSection( &m_cs );
}

/////////////////////////////////////////////////////////////////////////////

CFTS_MFX::~CFTS_MFX() 
{ 
	::DeleteCriticalSection( &m_cs );
	::InterlockedDecrement( &g_lComponents );
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// IMfxEventFilter

HRESULT CFTS_MFX::Connect( IUnknown* pContext )
{
 DECLARE_LOCAL(do_resets)
	// Here is our chance to acquire any IMFXEventFilter interfaces we need.
	pContext->QueryInterface( IID_IMfxTempoMap, (void**)&m_pTempoMap );
// 	pContext->QueryInterface( IID_IMfxDataQueue, (void**)&m_pDataQueue );
 /**/
#ifdef cl_do_resets
	pContext->QueryInterface( IID_IMfxInputPulse, (void**)&m_pInputPulse );
 if(m_pInputPulse)
 {
  if(do_resets)
  {
   prop.do_resets_again=1;
   m_pInputPulse->BeginPulse();
  }
  g_pInputPulse=m_pInputPulse;
 }
#endif
 /**/
 /**

	// IMfxInputCallback is necessary if the synth wants to push live MIDI
	// back to the host, but otherwise not mandated.
	pContext->QueryInterface( IID_IMfxInputCallback, (void**)&m_pInputCallback );
 **/

	// Since we're not retaining a copy of the original IUnknown*, be sure to
	// call Release() on it.
	pContext->Release();
	
	return S_OK;
}

//////////////////////////////////////////////////////////////////////
// IMfxInputPort

/**
STDMETHODIMP CFTS_MFX::SetInputCallback( IMfxInputCallback* pCallback )
{
	CAutoLock lock( m_pDXi );

	if (pCallback)
		pCallback->AddRef();
	if (m_pInputCallback)
		m_pInputCallback->Release();
	m_pInputCallback = pCallback;

	return S_OK;
}

////////////////////////////////////////////////////////////////////////////////

STDMETHODIMP CFTS_MFX::GetInputCallback( IMfxInputCallback** ppCallback )
{
	CAutoLock lock( m_pDXi );

	if (NULL == ppCallback)
		return E_POINTER;

	if (m_pInputCallback)
	{
		m_pInputCallback->AddRef();
		*ppCallback = m_pInputCallback;
		return S_OK;
	}
	else
	{
		*ppCallback = NULL;
		return S_FALSE;
	}
}
**/

/////////////////////////////////////////////////////////////////////////////

HRESULT CFTS_MFX::Disconnect()
{
	// Here we must release any interfaces we obtained in Connect()
	if (m_pTempoMap)
		m_pTempoMap->Release(), m_pTempoMap = NULL;
#ifdef cl_do_resets
	if (m_pInputPulse)
		m_pInputPulse->Release(), m_pInputPulse = g_pInputPulse = NULL;
#endif
	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////

char events_sent_multi_chann;
char shown_single_chann_barches;
#define MAX_MIDI_CHANNELS 16
int ipos_all_channs,ipos_note_off_all_channs;
char channels_found[MAX_MIDI_CHANNELS];
#define MAX_MIDI_NOTES 128
#define MAX_EVENTS_RECORD (MAX_MIDI_NOTES*MAX_MIDI_CHANNELS)
MfxEvent new_note_off_events[MAX_EVENTS_RECORD];
DWORD dwNoteOffEvents[MAX_EVENTS_RECORD];

#ifdef cl_MergeEvents
MfxEvent new_events_for_all_channs[MAX_EVENTS_RECORD];
MfxEvent new_note_off_events_for_all_channs[MAX_EVENTS_RECORD];
DWORD dwNoteOffEvents_for_all_channs[MAX_EVENTS_RECORD];
DWORD dwEvents_for_all_channs[MAX_EVENTS_RECORD];
#endif

void AddEventOff(MfxEvent note_off,DWORD dwnote_off)
{
 int i=0;
 for(i=0;i<MAX_EVENTS_RECORD;i++)
 if(dwNoteOffEvents[i]==0)
 {
  new_note_off_events[i]=note_off;
  dwNoteOffEvents[i]=dwnote_off;
  return;
 }
 new_note_off_events[0]=note_off;
 dwNoteOffEvents[0]=dwnote_off;
}

void InitVarsOnStartPlay(void)
{
 int i=0;
 for(i=0;i<MAX_EVENTS_RECORD;i++)
  dwNoteOffEvents[i]=0;
 events_sent_multi_chann=0;
 shown_single_chann_barches=0;
 chann_change_since_start_play=0;
 memset(&channels_found,0,sizeof(channels_found));
}

HRESULT CFTS_MFX::OnStart( LONG lTime, IMfxEventQueue* pqOut )
{
 InitVarsOnStartPlay();
	// NOTE: We are not allowed to retain the IMfxEventQueue pointer pqOut, nor
	// should we call Add() or Release() on it.
 if(hwndFTSMFX_is)
 {
  SendMessage(hwndFTSMFX_is,WM_START_MFX_RELAY,0,0);
 }
	return S_OK;
}

char FoundAnotherChann(int iCh)
{
 int i=0;
 for(i=0;i<MAX_MIDI_CHANNELS;i++)
  if(i!=iCh)
  if(channels_found[i])
   return 1;
 return 0;
}

#ifdef cl_MergeEvents
void MergeEvents(MfxEvent *events_all,MfxEvent *events
                ,DWORD  *dwall,DWORD *dwEvents
                ,int imax_events
                )
{
 if(events_all==NULL||dwall==NULL)
  return;
 if(dwEvents==NULL||events==NULL)
  return;
 // merge events with old events_all in time ordered sequence
 {
  int ipos_all=0;
  int ipos_new=0;
  int ipos_all_new=0;
  MfxEvent events_all_new[MAX_EVENTS_RECORD];
  DWORD dwall_new[MAX_EVENTS_RECORD];
  for(;;)
  {
   char do_ipos_new=0;
   if(ipos_new<MAX_EVENTS_RECORD&&ipos_new<imax_new_events)
   if(events[ipos_new]==0)
   {ipos_new++;continue;}
   if(ipos_all<MAX_EVENTS_RECORD)
   if(events[ipos_all]==0)
   {ipos_all++;continue;}
   if(ipos_all_new>=MAX_EVENTS_RECORD)
    break;
   if((ipos_new>=imax_new_events||ipos_new>=MAX_EVENTS_RECORD)&&ipos_all>=MAX_EVENTS_RECORD)
    break;
   if(ipos_new<MAX_EVENTS_RECORD)
   {
    if(events[ipos_new].m_lTime<events_all[ipos_all].m_lTime)
     do_ipos_new=1;
   }
   if(do_ipos_new)
   {
    dwall_new[ipos_all_new]=dwEvents[ipos_new];
    events_all_new[ipos_all_new]=events[ipos_new];
    ipos_new++;
   }
   else
   {
    dwall_new[ipos_all_new]=dwall[ipos_all];
    events_all_new[ipos_all_new]=events_all[ipos_all];
    ipos_all++;
   }
   ipos_all_new++;
  }
 }
 {
  int i=0;
  for(i=0;i<MAX_EVENTS_RECORD;i++)
  {
   dwall[i]=dwall_new[i];
   events_all[i]=events_all_new[i];
  }
 }
}
#endif

/////////////////////////////////////////////////////////////////////////////
#define MIDI_NOTE_ON 0x90
#define MIDI_NOTE_OFF 0x80
#define MIDI_CONTROLLER 0xb0
#define MIDI_SET_PATCH 0xC0
#define MIDI_CHANNEL_PRESSURE 0xd0
#define MIDI_AFTERTOUCH MIDI_CHANNEL_PRESSURE
#define MIDI_PITCH_BEND 0xE0
char *sz_short_msg(char *szBuffer,DWORD dw)
{
	WORD lw=LOWORD(dw),hw=HIWORD(dw);
	BYTE status=LOBYTE(lw),data1=HIBYTE(lw),data2=LOBYTE(hw);
 BYTE type=status&0xf0;
 BYTE chann=status&0x0f;
 int nvals=2;
 char szType[100];
 char *szdata1="data1 ",*szdata2="data2";
 switch(type)
 {
  case MIDI_NOTE_OFF:szdata1="";szdata2="vel.";strcpy(szType,"Note Off");break;
  case MIDI_NOTE_ON:szdata1="";szdata2="vel.";strcpy(szType,"Note On");break;
  case MIDI_SET_PATCH:szdata1="";nvals=1;strcpy(szType,"Patch");break;
  case MIDI_PITCH_BEND:strcpy(szType,"Pitch Bend");break;
  case MIDI_CONTROLLER:szdata1="";szdata2="val";strcpy(szType,"Controller");break;
  case MIDI_AFTERTOUCH:szdata1="";nvals=1;strcpy(szType,"Aftertouch");break;
  default:sprintf(szType,"Type %x",type);break;
 }
 if(nvals==1)
 sprintf(szBuffer,"%s %s%d",szType,szdata1,data1);
 else
 sprintf(szBuffer,"%s %s%d, %s %d",szType,szdata1,data1,szdata2,data2);
 return szBuffer;
}

void LogEventSentToFTS(FILE *fplog,char ignore_event_type,MfxEvent event,DWORD dwMsg)
{
 char szBuffer[STRMAX];
 int iCh=event.m_byChan;
REDO_LOG
 if(ignore_event_type)
 {
  if(fplog)fprintf(fplog," %s Original channel %d -> FTS\n"
                  ,sz_short_msg(szBuffer,dwMsg),iCh+1);
  return;
 }
 switch(event.m_eType)
 {
  case MfxEvent::ShortMsg:
   {
    if(fplog)fprintf(fplog," Short Msg %s Original channel %d -> FTS\n"
                    ,sz_short_msg(szBuffer,event.m_dwShortMsg),iCh+1);
   }
   break;
  case MfxEvent::Patch:if(fplog)fprintf(fplog," Patch %d Original channel %d -> FTS\n",event.m_byPatch,iCh+1);break;
  case MfxEvent::Note:if(fplog)fprintf(fplog," Note %d Original channel %d -> FTS\n",event.m_byKey,iCh+1);break;
  case MfxEvent::Wheel:if(fplog)fprintf(fplog," Pitch Bend %g cents Original channel %d -> FTS\n",200*(event.m_nVal)/(double)8192,iCh+1);break;
  case MfxEvent::Control:if(fplog)fprintf(fplog," Controller %d data %d Original channel %d -> FTS\n",event.m_byNum,event.m_byVal,iCh+1);break;
  case MfxEvent::ChanAft:if(fplog)fprintf(fplog," ChanAft %d Original channel %d -> FTS\n",event.m_byAmt,iCh+1);break;
   
 }
REDO_LOG}

HRESULT CFTS_MFX::OnEvents( LONG lTimeFrom, LONG lTimeThru, IMfxEventQueue* pqIn, IMfxEventQueue* pqOut )
{
	// NOTE: We are not allowed to retain the IMfxEventQueue pointers pqIn or
	// pqOut, nor should we call Add() or Release() on them.
	//
	// This method will likely make up the bulk of your MFX effect.
 char event_processed_in_fts=0;
 DECLARE_LOCAL(bypass_fts)
 DECLARE_LOCAL(fts_aftertouch)
 DECLARE_LOCAL(fts_controllers)
 DECLARE_LOCAL(fts_pitch_bends)
 DECLARE_LOCAL(fts_patches)
 DECLARE_LOCAL(fts_notes)
 DECLARE_LOCAL(fts_raw_data)
 DECLARE_LOCAL(fts_send_on_if_not_processed)
 DECLARE_LOCAL(fts_mute)
 DECLARE_LOCAL(fts_log_to_text)
 DECLARE_LOCAL(fts_skip_all_controllers)
 DECLARE_LOCAL(fts_choose_channel_to_relay)
 DECLARE_LOCAL_INT(fts_channel_to_relay)
 DECLARE_LOCAL(fts_relay_to_same_chann)
 DECLARE_LOCAL(fts_process_all_channels_up_to)
 DECLARE_LOCAL_INT(fts_channel_to_process_up_to)
 DECLARE_LOCAL(fts_verbose_log)
 DECLARE_LOCAL(fts_do_pitch_bend_delay)
 DECLARE_LOCAL_INT(fts_pitch_bend_delay)
 FILE *fplog=NULL;
 HWND	hwndFTSMFX=bypass_fts||fts_mute?0:GetFTSWindow();
	int nCount;
 double dfirst_time=0,dlast_time=0;
 char found_1st_time=0,found_last_time=0;
 char ready_to_process=1;
#ifdef DEBUG_FPDBG
 // MessageBox(g_hDlgPropPage,"OnEvents 1","FTS",0);
#endif
 if(fts_process_all_channels_up_to)
  ready_to_process=0;
 __try
 {
 if(fts_log_to_text)
 {
  char szDllPath[MAX_PATH];
  GetModulePath(szDllPath,MAX_PATH,"fts_mfx");
  MakePathFile(szLogPFile,szDllPath,"messages\\mfx_log.txt");
  MakeAnyNewWindowsFolders(szLogPFile);
  fplog=fopen(szLogPFile,started_log?"at":"wt");
#ifdef _DEBUG
  if(!started_log)
   MessageBox(g_hDlgPropPage,"Started log","debug",0);
#endif
  started_log=1;
 }
 if(fplog&&!log_header_up_to_date)
 AddLogHeader(fplog);
 REDO_LOG
 if(fts_mute)
 {
  if(fplog)fprintf(fplog,"Mute - skipping all events in this batch\n\n"
                   );
  return S_OK;
 }
	if (SUCCEEDED( pqIn->GetCount( &nCount ) ))
	{
  int ifirstCh=-1;
  int nMaxBuff=nCount*3;
  int ipos_note_off=0;
  DWORD *dwEvents=hwndFTSMFX?(DWORD *)malloc(sizeof(DWORD)*nMaxBuff):0;
  MfxEvent *new_events=hwndFTSMFX?(MfxEvent *)malloc(sizeof(MfxEvent)*nMaxBuff):0;
  int ipos=0;
  if(fplog&&fts_verbose_log)fprintf(fplog,"\n\nParsing events\nNumber of events to parse: %d\n"
                   ,nCount
                   );
		//loop through events
		for (int ix = 0; ix < nCount; ++ix)
		{
			MfxEvent event;
			//get each event
			if (SUCCEEDED( pqIn->GetAt( ix, &event ) ))
			{
    DWORD dw=0;
    char record_for_fts=1;
    int iStatus=0,iData1=0,iData2=0;
    int iCh=event.m_byChan;
    char *sztoprocess="";
    if(event.m_eType==MfxEvent::Patch)
     dw=dw;
   	if(m_pTempoMap)
    {
     if(ix==0)
     {
      found_1st_time=1;
      dfirst_time=m_pTempoMap->TicksToMsecs(event.m_lTime)/(double)1000;
     }
     else
     {
      found_last_time=1;
      dlast_time=m_pTempoMap->TicksToMsecs(event.m_lTime)/(double)1000;
     }
    }
    if(fplog&&fts_verbose_log)
   	if(m_pTempoMap)
     fprintf(fplog,"n %d\t time %#g\t ",ix,m_pTempoMap->TicksToMsecs(event.m_lTime)/(double)1000);
    if((fts_choose_channel_to_relay&&iCh!=fts_channel_to_relay-1)
      ||dwEvents==0||new_events==0
#ifdef cl_MergeEvents
      ||(fts_process_all_channels_up_to&&iCh>=fts_channel_to_process_up_to)
#endif
      )
    {
     if(fplog&&fts_verbose_log)fprintf(fplog,"\n"
                     );
     if(fplog&&fts_verbose_log)fprintf(fplog,fts_send_on_if_not_processed||bypass_fts?"(sent on unchanged)\n":"(skipped)\n");
	    if(fts_send_on_if_not_processed||bypass_fts)
     {
 	   	pqOut->Add( event );
      sztoprocess=" - sent on unchanged";
     }
     else
      sztoprocess=" - skipped";
     record_for_fts=0;
     if(!fplog)
      continue;
    }
    if(record_for_fts)
    {
     dwEvents[ipos]=0;
     new_events[ipos]=event;
    }
    // convert the eventS to dws if nec.
    switch(event.m_eType)
    {
     case MfxEvent::ShortMsg:
     case MfxEvent::Patch:
     case MfxEvent::Note:
     case MfxEvent::Wheel:
     case MfxEvent::Control:
     case MfxEvent::ChanAft:
      channels_found[iCh]=1;
      {
       char change_found=0;
       if(FoundAnotherChann(iCh))
        chann_change_since_start_play=1;
       if(ifirstCh>=0)
       if(ifirstCh>=0&&iCh!=ifirstCh)
       {
        events_sent_multi_chann=1;
        if(shown_single_chann_barches)
        {
         if(ifirstCh>=0&&iCh!=ifirstCh)
         {
          if(fplog&&fts_verbose_log)fprintf(fplog,"first chann found in stream %d, this one is channel %d - events sent several channels at a time\n"
                          ,ifirstCh,iCh
                          );
         }
         ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_ONLY_ONE_CHANNEL),SW_HIDE);
         ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_MULTI_CHANN),SW_HIDE);
         ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_SINGLE_CHANN),SW_SHOWNORMAL);
         shown_single_chann_barches=0;
        }
       }
       if(ifirstCh<0)
        ifirstCh=iCh;
      }
      if(fplog&&fts_verbose_log)fprintf(fplog,"chann %d\t ",iCh+1);
    }
    switch(event.m_eType)
    {
     case MfxEvent::ShortMsg:
      {
       char process=0;
       DWORD dw=event.m_dwShortMsg;
	      WORD lw=LOWORD(dw),hw=HIWORD(dw);
	      BYTE status=LOBYTE(lw),data1=HIBYTE(lw),data2=LOBYTE(hw);
       BYTE type=status&0xf0;
       BYTE chann=status&0x0f;
       char szType[100];
       if(fts_raw_data)
       switch(type)
        {
         case MIDI_NOTE_OFF:strcpy(szType,"Note Off");process=fts_notes;break;
         case MIDI_NOTE_ON:strcpy(szType,"Note On");process=fts_notes;break;
         case MIDI_SET_PATCH:strcpy(szType,"Patch");process=fts_patches;break;
         case MIDI_PITCH_BEND:strcpy(szType,"Pitch Bend");process=fts_pitch_bends;break;
         case MIDI_CONTROLLER:strcpy(szType,"Controller");process=fts_controllers;break;
         case MIDI_AFTERTOUCH:strcpy(szType,"Aftertouch");process=fts_aftertouch;break;
         default:sprintf(szType,"Type: %x",type);break;
        }
       if(process&&record_for_fts)
       {
        dwEvents[ipos]=event.m_dwShortMsg;
        ipos++;
        sztoprocess=" - for FTS";
       }
       {
        char change_found=0;
        if(FoundAnotherChann(iCh))
         chann_change_since_start_play=1;
        if(ifirstCh>=0)
        if(ifirstCh>=0&&iCh!=ifirstCh)
        {
         events_sent_multi_chann=1;
         if(shown_single_chann_barches)
         {
          if(ifirstCh>=0&&iCh!=ifirstCh)
          {
           if(fplog&&fts_verbose_log)fprintf(fplog,"first chann found in stream %d, this one is channel %d - events sent several channels at a time\n"
                           ,ifirstCh,iCh
                           );
          }
          ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_ONLY_ONE_CHANNEL),SW_HIDE);
          ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_MULTI_CHANN),SW_HIDE);
          ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_SINGLE_CHANN),SW_SHOWNORMAL);
          shown_single_chann_barches=0;
         }
        }
       }
       if(fplog&&fts_verbose_log)fprintf(fplog,"ShortMsg %s\t chann %d\t data1 %d\t data2 %d%s\n"
                        ,szType,chann+1,data1,data2,sztoprocess
                        );
      }
      continue;
#define MAKE_MSG(dwMessage,iStatus,iChannel,iData1,iData2)\
   			dwMessage = iStatus | iChannel | (iData1 << 8) | (iData2 << 16) ;
     case MfxEvent::Patch:
       if(fts_patches&&record_for_fts)
       {
        MAKE_MSG(dwEvents[ipos],MIDI_SET_PATCH,iCh,event.m_byPatch,0)
        ipos++;
        sztoprocess=" - for FTS";
       }
       if(fplog&&fts_verbose_log)fprintf(fplog,"Patch %d\t bank %d%s\n"
                                        ,event.m_byPatch,event.m_nBank,sztoprocess);
       continue;
     case MfxEvent::Note:
       if(fts_notes&&record_for_fts)
       {
        DWORD dwevent_off;
        MAKE_MSG(dwEvents[ipos],MIDI_NOTE_ON,iCh,event.m_byKey,event.m_byVel)
        ipos++;
        MAKE_MSG(dwevent_off,MIDI_NOTE_OFF,iCh,event.m_byKey,event.m_byVelOff)
        AddEventOff(event,dwevent_off);
        ipos_note_off++;
        sztoprocess=" - for FTS as note on / off pair";
       }
       if(fplog&&fts_verbose_log)fprintf(fplog,"Note %d\t Vel %d\t "
                        ,event.m_byKey,event.m_byVel
                        );
       if(fplog&&fts_verbose_log)
   	   if(m_pTempoMap)
        fprintf(fplog,"duration %#g",m_pTempoMap->TicksToMsecs(event.m_dwDuration)/(double)1000);
       if(fplog&&fts_verbose_log)fprintf(fplog,"%s\n",sztoprocess
                       );
       continue;
     case MfxEvent::Wheel:
      {
       int ipb=(int)(event.m_nVal+8192);
       if(fts_pitch_bends&&record_for_fts)
       {
        MAKE_MSG(dwEvents[ipos],MIDI_PITCH_BEND,iCh,ipb&0x7F,ipb>>7)
        sztoprocess=" - for FTS";
        ipos++;
       }
       if(fplog&&fts_verbose_log)fprintf(fplog,"- Pitch Bend\t %d (%d) = %#g cents%s\n"
                       ,ipb,ipb-8192
                       ,200*(ipb-8192)/(double)8192
                       ,sztoprocess
                       );
      }
      continue;
     case MfxEvent::Control:
      if(fts_controllers&&record_for_fts)
      {
       MAKE_MSG(dwEvents[ipos],MIDI_CONTROLLER,iCh,event.m_byNum,event.m_byVal)
       sztoprocess=" - for FTS";
       ipos++;
      }
     if(fplog&&fts_verbose_log)fprintf(fplog,"- Controller %d\t value %d%s\n"
                                      ,event.m_byNum,event.m_byVal
                                      ,sztoprocess
                     );
      continue;
     case MfxEvent::ChanAft:
      if(fts_aftertouch&&record_for_fts)
      {
       MAKE_MSG(dwEvents[ipos],MIDI_AFTERTOUCH,iCh,event.m_byAmt,0)
       sztoprocess=" - for FTS";
       ipos++;
      }
      if(fplog&&fts_verbose_log)fprintf(fplog,"- Aftertouch %d%s\n",event.m_byAmt,sztoprocess
                      );
      continue;
     default:
      // dwEvents[ipos]=0;
      if(fplog&&fts_verbose_log)fprintf(fplog,"%s\n",sztoprocess
        );
  				// ipos++;
    }
   }
  }// ix
REDO_LOG
#ifdef cl_MergeEvents
  if(fts_process_all_channels_up_to)
  {
   #ifdef DEBUG_FPDBG
    {
     FILE *fpdbg=fopen("i:\\log.txt","at");
     fprintf(fpdbg,"\n\nMergeEvents new_events\n"
            );
     fclose(fpdbg);
    }
   #endif
   // merge these events with the ones already found
   MergeEvents(new_events_for_all_channs,new_events
              ,dwEvents_for_all_channs,dwEvents
              ,ipos
              );
   #ifdef DEBUG_FPDBG
    {
     FILE *fpdbg=fopen("i:\\log.txt","at");
     fprintf(fpdbg,"\n\nMergeEvents new_note_off_events\n"
            );
     fclose(fpdbg);
    }
   #endif
   MergeEvents(new_note_off_events_for_all_channs,new_note_off_events
              ,dwNoteOffEvents_for_all_channs,dwNoteOffEvents
              ,ipos_note_off
              );
   #ifdef DEBUG_FPDBG
    {
     FILE *fpdbg=fopen("i:\\log.txt","at");
     fprintf(fpdbg,"\n\nMergeEvents after both - ready_to_process %d\n"
            ,ready_to_process
            );
     fclose(fpdbg);
    }
   #endif
   ready_to_process=1;
   if(fplog&&fts_verbose_log)fprintf(fplog,"\n\nchannels found so far: (max channel to process %d)\n",fts_channel_to_process_up_to);
   int ii=0;
   for(ii=0;ii<fts_channel_to_process_up_to;ii++)
    if(channels_found[ii])
    {
     if(fplog&&fts_verbose_log)fprintf(fplog,"%d ",ii+1);
    }
    else
     ready_to_process=0;
   if(fplog&&fts_verbose_log)fprintf(fplog,"\n\n");
  }
#endif
REDO_LOG
  if(ready_to_process)
  if(dwEvents&&new_events)
  {
   int ipos_reached=ipos;
   int ipos_note_off_reached=ipos_note_off;
   DWORD *dwEventsIs=dwEvents;
   DWORD *dwNoteOffEventsIs=dwNoteOffEvents;
   MfxEvent *new_events_is=new_events;
   MfxEvent *new_note_off_events_is=new_note_off_events;
   if(ipos_reached>0||ipos_note_off_reached>0||fts_verbose_log)
   {
    if(fplog)fprintf(fplog,"\n\nProcessing events\n");
    #ifdef DEBUG_FPDBG
     {
      FILE *fpdbg=fopen("i:\\log.txt","at");
      fprintf(fpdbg,"aa "
             );
      fclose(fpdbg);
     }
    #endif
    if(fplog)fprintf(fplog,"Number of events to relay via FTS %d\n",ipos_reached);
    if(fplog)if(ipos_note_off_reached)fprintf(fplog,"Number of extra note off events to relay via FTS %d\n",ipos_note_off_reached);
   }
   #ifdef DEBUG_FPDBG
    {
     FILE *fpdbg=fopen("i:\\log.txt","at");
     fprintf(fpdbg,"bb "
            );
     fclose(fpdbg);
    }
   #endif
#ifdef cl_MergeEvents
   if(fts_process_all_channels_up_to)
   {
    dwEventsIs=dwEvents_for_all_channs;
    dwNoteOffEventsIs=dwNoteOffEvents_for_all_channs;
    new_events_is=new_events;
    new_note_off_events_is=new_note_off_events_for_all_channs;
    ipos_reached=ipos_all_channs;
    ipos_note_off_reached=ipos_note_off_all_channs;
    #ifdef DEBUG_FPDBG
     {
      FILE *fpdbg=fopen("i:\\log.txt","at");
      fprintf(fpdbg,"\n\nipos_all_channs %d ipos_note_off_reached %d\n"
             ,ipos_all_channs,ipos_note_off_reached
             );
      fclose(fpdbg);
     }
    #endif
   }
#endif
   if(dwEventsIs&&dwNoteOffEventsIs&&new_events_is&&new_note_off_events_is)
   {
      EVENTS_IN_OUT *pEventsInOut=NULL;
    // events are linearly ordered in time.
    ipos_note_off=0;
    if(hwndFTSMFX&&ipos_reached>0)
    {
     SendMessage(hwndFTSMFX,WM_DO_MFX_SAME_CHANN,fts_relay_to_same_chann?1:0,0);
    }
    for(ipos=0;ipos<ipos_reached;ipos++)
    {
 			 MfxEvent event=new_events_is[ipos];
     if(fplog)
   	 if(m_pTempoMap)
      fprintf(fplog,"time %#g\t ",m_pTempoMap->TicksToMsecs(event.m_lTime)/(double)1000);
     if(dwEventsIs[ipos]==0)
     {
	     if(fts_send_on_if_not_processed)
       pqOut->Add(event);
      if(fplog)fprintf(fplog,fts_send_on_if_not_processed?"(sent on unchanged)\n":"(skipped)\n");
     }
     else
     {
      int iCh=event.m_byChan;
      int iwhat_to_do=0;
      char do_break=0;
      DWORD dwEventToSend=0;
      // send on any note offs to get up to date with the time.
      // Just ignore the replies - this is just so that FTS will record
      // them at the correct times.
      for(;;)
      {
       char found_note_off=0;
       for(ipos_note_off=0;ipos_note_off<MAX_EVENTS_RECORD;ipos_note_off++)
       if(dwNoteOffEventsIs[ipos_note_off])
       {
        long lnew_note_off_time=new_note_off_events_is[ipos_note_off].m_lTime
               +(long)new_note_off_events_is[ipos_note_off].m_dwDuration;
        if(event.m_lTime>=lnew_note_off_time)
        {
         iwhat_to_do=MFX_SKIP_NOTE_OFFS;
         dwEventToSend=dwNoteOffEventsIs[ipos_note_off];
         if(fplog)
         fprintf(fplog," Note expired at %g: ",m_pTempoMap->TicksToMsecs(lnew_note_off_time)/(double)1000);
         LogEventSentToFTS(fplog,1/*ignore_event_type*/,event,dwEventToSend);
         dwNoteOffEventsIs[ipos_note_off]=0;
         found_note_off=1;
         break;
        }
       }
       if(do_break&&!found_note_off)
        break;
       if(!found_note_off)
       {
        iwhat_to_do=0;
        dwEventToSend=dwEventsIs[ipos];
        /**
        {
         DWORD dw=dwEventToSend;\
	        WORD lw=LOWORD(dw);
	        WORD hw=HIWORD(dw);
         BYTE note=HIBYTE(lw);
         BYTE vel=LOBYTE(hw);
	        BYTE status=LOBYTE(lw);\
         BYTE type=status&0xf0;\
         BYTE chann=status&0x0f;
         if(status==MIDI_NOTE_ON&&vel==0)
         dwEventToSend=MAKELONG(127,MAKEWORD(note,MIDI_NOTE_OFF|chann)); 
        }
        **/
        LogEventSentToFTS(fplog,0/*ignore_event_type*/,event,dwEventToSend);
        do_break=1;
       }
       if(dwEventToSend==0)
        continue;
#ifdef _DEBUG
     // continue;
#endif
       {
        if(pEventsInOut)
         FreeFileMapMem(pEventsInOut);   pEventsInOut=NULL;      
        ConstructFileMapping();
        GetFileMapMem(&pEventsInOut);
	       if(pEventsInOut!=NULL)
        {
         pEventsInOut->iversion=FTS_MFX_RELAY_VERSION_1;
         pEventsInOut->iwhat_to_do=iwhat_to_do;
         pEventsInOut->nOutputVals=0;
         pEventsInOut->dwInput=dwEventToSend;
         SendMessage(hwndFTSMFX,WM_DO_MFX_RELAY,0,0);
         event_processed_in_fts=1;
        }
        if(fplog&&fts_verbose_log)if(pEventsInOut==NULL)
									 fprintf(fplog,"\nBUG - pEventsInOut==NULL\n");
        if(fplog&&fts_verbose_log)
        if(pEventsInOut)
								{
										 fprintf(fplog,"To add: %d events\n",pEventsInOut->nOutputVals);
										 fprintf(fplog,"pEventsInOut->dwInput %u\n",pEventsInOut->dwInput);
										 fprintf(fplog,"pEventsInOut->dwOutput %u\n",pEventsInOut->dwOutput);
								}
        if(pEventsInOut&&pEventsInOut->nOutputVals>0)
        {
         int jpos=0;
         int nOutputVals=pEventsInOut->nOutputVals;
         DWORD *dwOutput=pEventsInOut->dwOutput;
   #ifdef DEBUG_MSGS
         {
          char szmsg[1024];
          sprintf(szmsg,"dwEventsIs[ipos] %lu, nOutputVals %d\n",dwEventsIs[ipos],nOutputVals);
          MessageBox(g_hDlgPropPage,szmsg,"debug",0);
         }
   #endif
         if(pEventsInOut->nSysexes>0)
         {
          DWORD dwStartSysex=0;
          BYTE *byMFXSysexVals=pEventsInOut->byMFXSysexVals;
          int nsysexes_relayed=0;
          if(fplog)fprintf(fplog,"\n%d sysexes to relay\n",pEventsInOut->nSysexes);
          for(jpos=0;jpos<pEventsInOut->nSysexes;jpos++)
		        {
           DWORD dwLengthSysex=pEventsInOut->dwSysexLength[jpos];
		         IMfxEventQueue2 *pq2=NULL;
           if(dwStartSysex+dwLengthSysex>MAX_byMFXSysexVals)
            break;
			        if(!FAILED(pqIn->QueryInterface(IID_IMfxEventQueue2,(void**)&pq2)))
           {
 			        IMfxBufferFactory *pBufFact=NULL;
		 	        if(SUCCEEDED(pq2->GetBufferFactory(&pBufFact)))
			         {
             MFX_HBUFFER hbufNew=NULL;
             if(S_OK==pBufFact->Create(dwLengthSysex,&hbufNew))
             {
				          void* pv=NULL;
				          DWORD cb=0;
				          if(SUCCEEDED(pBufFact->GetPointer(hbufNew,&pv,&cb)))
              {
				       	   BYTE* pby=static_cast<BYTE*>(pv); 
               memcpy(pby,&byMFXSysexVals[dwStartSysex],dwLengthSysex);
               event.m_eType=MfxEvent::Sysx;
               event.m_hBuffer=hbufNew;
               if(fplog)fprintf(fplog," sysex\t ");
               nsysexes_relayed++;
	   	          pqOut->Add(event);
              }
             }
				         pBufFact->Release();
			         }
			         pq2->Release();
		         }
           dwStartSysex+=dwLengthSysex;
          }
          if(fplog)fprintf(fplog,"\n%d sysexes relayed successfully\n",nsysexes_relayed);
         }
         for(jpos=0;jpos<nOutputVals;jpos++)
         {
          DWORD dwMidiMessage=dwOutput[jpos];
          // translate back from short msg
	         WORD lw=LOWORD(dwMidiMessage),hw=HIWORD(dwMidiMessage);
	         BYTE status=LOBYTE(lw),data1=HIBYTE(lw),data2=LOBYTE(hw);
          BYTE type=status&0xf0;
          BYTE chann=status&0x0f;
          event.m_byChan=chann;
          /**
          switch(type)
          {
           case MIDI_NOTE_ON:
           case MIDI_SET_PATCH:
           case MIDI_PITCH_BEND:
           case MIDI_CONTROLLER:
           case MIDI_AFTERTOUCH:
          }
          **/
          switch(type)
          {
          case MIDI_NOTE_OFF:// note offs only relayed from FTS if needed 
                             // to free channels for pitch bends
           event.m_eType=MfxEvent::ShortMsg;
           event.m_dwShortMsg=dwMidiMessage;
           if(fplog)fprintf(fplog," \tchann %d\t Note off %d\t Vel %d\t (note cut short because chann is needed for pitch bend)"
                            ,chann+1,event.m_byKey,event.m_byVel
                            );
           break;
          case MIDI_NOTE_ON:
           event.m_eType=MfxEvent::Note;
           event.m_byKey=data1;
           event.m_byVel=data2;
   #define SET_IF_SAME_TYPE(type,var)\
           event.var=new_events_is[ipos].m_eType==MfxEvent::type?new_events_is[ipos].var:0;
           SET_IF_SAME_TYPE(Note,m_dwDuration)
           if(fplog)fprintf(fplog,"\tchann %d\t Note %d\t Vel %d\t "
                            ,chann+1,event.m_byKey,event.m_byVel 
                            );
           if(fplog)
   	       if(m_pTempoMap)
            fprintf(fplog,"duration %#g\t ",m_pTempoMap->TicksToMsecs(event.m_dwDuration)/(double)1000);
          break;
          case MIDI_SET_PATCH:
           event.m_eType=MfxEvent::Patch;
           event.m_byPatch=data1;
           SET_IF_SAME_TYPE(Patch,m_byBankSelMethod)
           SET_IF_SAME_TYPE(Patch,m_nBank)
           if(fplog&&(fts_verbose_log||fts_patches))fprintf(fplog,"\tchann %d\t Patch %d\t bank %d\t ",chann+1,data1,event.m_nBank);
           if(!fts_patches)
           {
            if(fplog)fprintf(fplog," - skipped\n");
            continue;
           }
           break;
          case MIDI_PITCH_BEND:
           event.m_eType=MfxEvent::Wheel;
           {
 			        int ipitch_bend=data1|data2<<7;
            event.m_nVal=ipitch_bend-8192;
            if(fplog)
            {
             if(fts_verbose_log)
             fprintf(fplog," \tchann %d\t Pitch Bend %d (%d) = %#g cents",chann+1,ipitch_bend,ipitch_bend-8192
                            ,200*(ipitch_bend-8192)/(double)8192
                            );
             else
             fprintf(fplog," \tchann %d\t Pitch Bend %#g cents",chann+1
                            ,200*(ipitch_bend-8192)/(double)8192
                            );
            }
            break;
           }
          case MIDI_CONTROLLER:
           event.m_eType=MfxEvent::Control;
           event.m_byNum=data1;
           event.m_byVal=data2;
           if(fplog&&(fts_verbose_log||fts_skip_all_controllers==0))fprintf(fplog," \tchann %d\t Controller %d\t value %d\t ",chann+1,data1,data2
                           );
           if(fts_skip_all_controllers)
           {
            if(fplog&&fts_verbose_log)fprintf(fplog," - skipped\n");
            continue;
           }
           break;
          case MIDI_AFTERTOUCH:
           event.m_eType=MfxEvent::ChanAft;
           event.m_byAmt=data1;
           if(fplog&&(fts_verbose_log||fts_aftertouch))fprintf(fplog," \tchann %d\t Aftertouch %d\t ",chann+1,data1
                           );
           if(fts_aftertouch==0)
            continue;
           break;
          default:
           // if(fts_send_on_if_not_processed==0)
           {
            if(fplog&&fts_verbose_log)fprintf(fplog," \tchann %d type %x data 1 %d data 2 %d\t - skipped"
                                             ,chann+1,type,data1,data2
                            );
            continue;
           }
          }
          pqOut->Add(event);
          if(fplog)fprintf(fplog,"\t - added\n");
         }// jpos
        }
        if(pEventsInOut)
         FreeFileMapMem(pEventsInOut); pEventsInOut=NULL;        
       }// ConstructFileMapping
      }
     }
    }// ipos
    if(pEventsInOut)
     FreeFileMapMem(pEventsInOut); pEventsInOut=NULL;      
   }
		}
 if(hwndFTSMFX&&event_processed_in_fts&&found_1st_time&&found_last_time)
   SendMessage(hwndFTSMFX,WM_ADD_TO_PLAYING_TIME,(LPARAM)(dlast_time-dfirst_time),0);
  #ifdef DEBUG_FPDBG
   {
    FILE *fpdbg=fopen("i:\\log.txt","at");
    fprintf(fpdbg,"xx \n"
           );
    fclose(fpdbg);
   }
  #endif
  FREE_IF_NON_ZERO(dwEvents)
  FREE_IF_NON_ZERO(new_events)
  #ifdef DEBUG_FPDBG
   {
    FILE *fpdbg=fopen("i:\\log.txt","at");
    fprintf(fpdbg,"yy \n"
           );
    fclose(fpdbg);
   }
  #endif
  if(ready_to_process)
  {
   int ii=0;
   for(ii=0;ii<MAX_MIDI_CHANNELS;ii++)
    channels_found[ii]=0;
  #ifdef DEBUG_FPDBG
   {
    FILE *fpdbg=fopen("i:\\log.txt","at");
    fprintf(fpdbg,"zz \n"
           );
    fclose(fpdbg);
   }
  #endif
  }
  if(events_sent_multi_chann==0)
  if(nCount>1)
  {
   if(fplog)fprintf(fplog,"So far, no changes of channel in the middle of a batch of events\n");
   if(!shown_single_chann_barches)
   {
    if(chann_change_since_start_play)
    {
     ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_SINGLE_CHANN),SW_SHOWNORMAL);
     ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_ONLY_ONE_CHANNEL),SW_HIDE);
    }
    else
    {
     ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_ONLY_ONE_CHANNEL),SW_SHOWNORMAL);
     ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_SINGLE_CHANN),SW_HIDE);
    }
    ShowWindow(GetDlgItem(g_hDlgPropPage,IDC_STATIC_BATCHES_MULTI_CHANN),SW_HIDE);
    shown_single_chann_barches=1;
   }
  }
  if(fplog)fclose(fplog);
  #ifdef DEBUG_FPDBG
   {
    FILE *fpdbg=fopen("i:\\log.txt","at");
    fprintf(fpdbg,"ww \n"
           );
    fclose(fpdbg);
   }
  #endif
	}
 }
 __except(TRUE)
 {
#ifdef _DEBUG
  MessageBox(NULL,"Exception occurred","FTS MFX",MB_ICONEXCLAMATION);
#endif
 }
	return S_OK;
}


/////////////////////////////////////////////////////////////////////////////

HRESULT CFTS_MFX::OnLoop( LONG lTimeRestart, LONG lTimeStop, IMfxEventQueue* pqOut )
{
	// NOTE: We are not allowed to retain the IMfxEventQueue pointer pqOut, nor
	// should we call Add() or Release() on it.

	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////

HRESULT CFTS_MFX::OnStop( LONG lTime, IMfxEventQueue* pqOut )
{
	// NOTE: We are not allowed to retain the IMfxEventQueue pointer pqOut, nor
	// should we call Add() or Release() on it.

	return S_OK;
}

// IMfxInputCallback::OnEvent could be used to play fractal tune
// from FTS in CakeWalk
void AddFTSData(char *szName,FILE *fplog,HWND hwndFTSMFX,IMfxDataQueue*  pqOut
                ,IMfxTempoMap* m_pTempoMap
                ,MfxData mfxdata,DWORD dwData,WORD message_for_FTS
                ,char *pevent_processed_in_fts
                )
{
 EVENTS_IN_OUT *pEventsInOut=NULL;
 DECLARE_LOCAL(fts_skip_all_controllers)
 DECLARE_LOCAL(fts_verbose_log)
 DECLARE_LOCAL(fts_do_pitch_bend_delay)
 DECLARE_LOCAL_INT(fts_pitch_bend_delay)
 DECLARE_LOCAL(fts_patches)
 DECLARE_LOCAL(fts_aftertouch)
 DECLARE_LOCAL(fts_notes)
 DECLARE_LOCAL(fts_pitch_bends)
 char needs_pb_delay[MAX_MIDI_CHANNELS];
 int i=0;
 for(i=0;i<MAX_MIDI_CHANNELS;i++)
  needs_pb_delay[i]=0;
 ConstructFileMapping();
 GetFileMapMem(&pEventsInOut);
	if(pEventsInOut!=NULL)
 {
  time_t tstart=clock();;
  static double dlat_max;
  pEventsInOut->iversion=FTS_MFX_RELAY_VERSION_1;
  pEventsInOut->iwhat_to_do=MFX_PROCESS_MIDI_KBD;
  pEventsInOut->nOutputVals=0;
  pEventsInOut->dwInput=dwData;
  SendMessage(hwndFTSMFX,message_for_FTS,0,0);
  (*pevent_processed_in_fts)=1;
  dlat_max=max(dlat_max,(double)(clock()-tstart)/CLOCKS_PER_SEC);
  dlat_max=dlat_max;
 }
 if(pEventsInOut&&pEventsInOut->nOutputVals>0)
 {
  int jpos=0;
  int nOutputVals=pEventsInOut->nOutputVals;
  DWORD *dwOutput=pEventsInOut->dwOutput;
  for(jpos=0;jpos<nOutputVals;jpos++)
  {
   DWORD dw=dwOutput[jpos];\
	  WORD lw=LOWORD(dw);
	  BYTE status=LOBYTE(lw);\
   BYTE type=status&0xf0;\
   BYTE chann=status&0x0f;\
   MfxData mfxdata2=mfxdata;
   if(needs_pb_delay[chann])
    mfxdata2.m_lTime+=fts_pitch_bend_delay;
   if(fts_do_pitch_bend_delay)
   if(type==MIDI_PITCH_BEND)
    needs_pb_delay[chann]=1;
   mfxdata2.m_dwData=dw;
   switch(type)
   {
   case MIDI_PITCH_BEND:
    if(!fts_pitch_bends)
    {
     if(fplog&&fts_verbose_log)fprintf(fplog,"%s: Pitch bend - skipped\n",szName);
     continue;
    }
    break;
   case MIDI_NOTE_ON:
   case MIDI_NOTE_OFF:
    if(!fts_notes)
    {
     if(fplog&&fts_verbose_log)fprintf(fplog,"%s: Note on or off - skipped\n",szName);
     continue;
    }
    break;
   case MIDI_SET_PATCH:
    if(!fts_patches)
    {
     if(fplog&&fts_verbose_log)fprintf(fplog,"%s: Patch - skipped\n",szName);
     continue;
    }
    break;
   case MIDI_CONTROLLER:
    if(fts_skip_all_controllers)
    {
     if(fplog&&fts_verbose_log)fprintf(fplog,"%s: Controller - skipped\n",szName);
     continue;
    }
    break;
   case MIDI_AFTERTOUCH:
    if(!fts_aftertouch)
    {
     if(fplog&&fts_verbose_log)fprintf(fplog,"%s: Aftertouch - skipped\n",szName);
     continue;
    }
    break;
   default:
    break;
   }
#define AddToLog(szName,mfxdatais,dwData)\
     {\
      char process=0;\
      DWORD dw=dwData;\
	     WORD lw=LOWORD(dw),hw=HIWORD(dw);\
	     BYTE status=LOBYTE(lw),data1=HIBYTE(lw),data2=LOBYTE(hw);\
      BYTE type=status&0xf0;\
      BYTE chann=status&0x0f;\
      char szType[100];\
      char *szdata1="data1",*szdata2="data2";\
      switch(type)\
       {\
        case MIDI_NOTE_OFF:strcpy(szType,"Note Off");szdata1="note";szdata2="vel";break;\
        case MIDI_NOTE_ON:strcpy(szType,"Note On");szdata1="note";szdata2="vel";break;\
        case MIDI_SET_PATCH:strcpy(szType,"Patch");szdata1="patch";break;\
        case MIDI_PITCH_BEND:strcpy(szType,"Pitch Bend");szdata1="coarse";szdata2="fine";break;\
        case MIDI_CONTROLLER:strcpy(szType,"Controller");szdata1="ctrlr";szdata2="value";break;\
        case MIDI_AFTERTOUCH:strcpy(szType,"Aftertouch");szdata1="note";szdata2="value";;break;\
        default:sprintf(szType,"%x",type);break;\
       }\
      if(fplog)\
      {\
       fprintf(fplog,"\n%s:",szName);\
      	if(m_pTempoMap)\
        fprintf(fplog,"time %#g\t ",m_pTempoMap->TicksToMsecs(mfxdatais.m_lTime)/(double)1000);\
       if(type==MIDI_PITCH_BEND)\
       {\
        int ipitch_bend=data1|(data2<<7);\
        fprintf(fplog,"%s: %s\t chann %d\t pitch bend %d (%d)"\
               ,szName,szType,chann+1,ipitch_bend\
               ,ipitch_bend-8192\
               );\
        fprintf(fplog," = %#g cents"\
               ,200*(ipitch_bend-8192)/(double)8192\
               );\
       }\
       else\
       fprintf(fplog,"%s: %s\t chann %d\t %s %d\t %s %d%s"\
              ,szName,szType,chann+1,szdata1,data1,szdata2,data2\
              ,type==MIDI_NOTE_ON&&data2==0?" (= note off)":""\
              );\
      }\
     }
   AddToLog("Output",mfxdata2,dwOutput[jpos])
   pqOut->Add(mfxdata2);
  }
 }
 if(pEventsInOut)
  FreeFileMapMem(pEventsInOut);       
}

/////////////////////////////////////////////////////////////////////////////

HRESULT CFTS_MFX::OnInput( IMfxDataQueue* pqIn, IMfxDataQueue* pqOut )
{
	int nCount;
 double dfirst_time=0,dlast_time=0;
 char found_1st_time=0,found_last_time=0;
 char event_processed_in_fts=0;
 __try
 {
 FILE *fplog=NULL;
 DECLARE_LOCAL(fts_midi_in_data)
 DECLARE_LOCAL(fts_mute_midi_in)
 DECLARE_LOCAL(fts_mute)
 DECLARE_LOCAL(bypass_fts)
 DECLARE_LOCAL(fts_skip_all_controllers)
 DECLARE_LOCAL(fts_aftertouch)
 DECLARE_LOCAL(fts_controllers)
 DECLARE_LOCAL(fts_pitch_bends)
 DECLARE_LOCAL(fts_patches)
 DECLARE_LOCAL(fts_notes)
 DECLARE_LOCAL(fts_raw_data)
 DECLARE_LOCAL(fts_send_on_if_not_processed)
 DECLARE_LOCAL(fts_log_to_text)
 DECLARE_LOCAL(fts_verbose_log)
 DECLARE_LOCAL(fts_choose_channel_to_relay)
 DECLARE_LOCAL_INT(fts_channel_to_relay)
 DECLARE_LOCAL(fts_do_pitch_bend_delay)
 DECLARE_LOCAL_INT(fts_pitch_bend_delay)
 DECLARE_LOCAL(fts_process_all_channels_up_to)
 DECLARE_LOCAL_INT(fts_channel_to_process_up_to)
 DECLARE_LOCAL(do_resets)
 DECLARE_LOCAL(do_resets_again)
 HWND	hwndFTSMFX=bypass_fts||fts_mute?0:GetFTSWindow();
	// This gives us a chance to process live MIDI input
#ifdef _DEBUG
 // MessageBox(g_hDlgPropPage,"OnInput 1","FTS",0);
#endif
#ifdef cl_do_resets
 if(fts_log_to_text)
 {
  char szDllPath[MAX_PATH];
  GetModulePath(szDllPath,MAX_PATH,"fts_mfx");
  MakePathFile(szLogPFile,szDllPath,"messages\\mfx_log.txt");
  MakeAnyNewWindowsFolders(szLogPFile);
  fplog=fopen(szLogPFile,started_log?"at":"wt");
#ifdef _DEBUG
  if(!started_log)
  started_log=started_log;//MessageBox(g_hDlgPropPage,"New log","debug",0);
#endif
  started_log=1;
 }
 if(fplog&&!log_header_up_to_date)
 {
  AddLogHeader(fplog);
 }
 if(g_pInputPulse&&do_all_sound_off)
 {
  MfxData mfxdata;
  DWORD   dwData=0;
  if(fplog)
   fprintf(fplog,"Reset all controllers and all sound off in all channels\n\n");
  mfxdata.m_lTime=0;
#define ADD_EVENT(time,chann,type,data1,data2)\
  dwData=MAKELONG(MAKEWORD((type)|(chann),data1),MAKEWORD(data2,0));\
  mfxdata.m_lTime=time;\
  mfxdata.m_dwData=dwData;\
  AddToLog("Output",mfxdata,dwData)\
  pqOut->Add(mfxdata);
#define ALL_SOUND_OFF 120
#define RESET_ALL_CONTROLLERS 0x79
  {
   int i=0;
   for(i=0;i<MAX_MIDI_CHANNELS;i++)
   {
    ADD_EVENT(0,i,MIDI_CONTROLLER,ALL_SOUND_OFF,0)
    ADD_EVENT(0,i,MIDI_CONTROLLER,RESET_ALL_CONTROLLERS,0)
    ADD_EVENT(0,i,MIDI_SET_PATCH,0,0)
   }
  }
  do_all_sound_off=0;
  g_pInputPulse->EndPulse();
 }
 if(g_pInputPulse&&do_pb_test)
 {
  MfxData mfxdata;
  DWORD   dwData=0;
  int ipitch_bend=8192;
  if(fplog)
   fprintf(fplog,"Pitch bend test\n\n");
  mfxdata.m_lTime=0;
  ipitch_bend=8192;
#define ADD_EVENT(time,chann,type,data1,data2)\
  dwData=MAKELONG(MAKEWORD((type)|(chann),data1),MAKEWORD(data2,0));\
  mfxdata.m_lTime=time;\
  mfxdata.m_dwData=dwData;\
  AddToLog("Output",mfxdata,dwData)\
  pqOut->Add(mfxdata);
  ADD_EVENT(0,0,MIDI_PITCH_BEND,ipitch_bend & 0x7F,ipitch_bend >> 7)
  ADD_EVENT(0,0,MIDI_NOTE_ON,60,127)
  ADD_EVENT(1000,0,MIDI_NOTE_ON,60,0)
  ipitch_bend=8192+8192/2;
  ADD_EVENT(1000,0,MIDI_PITCH_BEND,ipitch_bend & 0x7F,ipitch_bend >> 7)
  ADD_EVENT(1000,0,MIDI_NOTE_ON,59,127)
  ADD_EVENT(2000,0,MIDI_NOTE_ON,59,0)
  do_pb_test=0;
  g_pInputPulse->EndPulse();
 }
 else
 if(g_pInputPulse&&do_resets_again)
 {
  char temp_event_processed_in_fts=0;
  DWORD   dwData=0;
  MfxData mfxdata;
#define LOG_VAL(val,sz) {if(val)fprintf(fplog,"%s ",sz);}
  if(fplog)
   fprintf(fplog,"Resets (pitch bend range etc)\n\n");
  mfxdata.m_lTime=0;
  g_pInputPulse->EndPulse();
  prop.do_resets_again=0;
  UpdateChecks();
  InitVarsOnStartPlay();
  if(hwndFTSMFX)
  {
   time_t tstart=clock();;
   static double dlat_max;
   AddFTSData("resets",fplog,hwndFTSMFX,pqOut,m_pTempoMap,mfxdata,dwData,WM_RESET_MFX_RELAY,&temp_event_processed_in_fts);
   dlat_max=max(dlat_max,(double)(clock()-tstart)/CLOCKS_PER_SEC);
   dlat_max=dlat_max;
  }
 }
#endif
 if(fts_midi_in_data==0||fts_mute)
 {
 	return S_OK;
	}
 else
	if (SUCCEEDED( pqIn->GetCount( &nCount ) ))
	{
#ifdef _DEBUG
 // MessageBox(g_hDlgPropPage,"OnInput 2","FTS",0);
#endif
  if(fplog)fprintf(fplog,"\n\n\nMidi Input - %d events received to process\n"\
                   ,nCount\
                   );\
  if(hwndFTSMFX||bypass_fts)
		for (int ix = 0; ix < nCount; ++ix)
		{
			MfxData mfxdata;
			if (SUCCEEDED( pqIn->GetAt( ix, &mfxdata ) ))
			{
    DWORD dwData=mfxdata.m_dwData;
    BYTE status=LOBYTE(LOWORD(dwData));
    char process=0;
    if(fplog)fprintf(fplog,"\n");\
    AddToLog("Input",mfxdata,dwData)
    if(!bypass_fts)
    if(fts_choose_channel_to_relay&&(status&0x0f)!=fts_channel_to_relay)
    {
     if(fplog)fprintf(fplog," - Channel skipped\n"\
                      );\
     continue;
    }
   	if(m_pTempoMap)
    {
     if(ix==0)
     {
      found_1st_time=1;
      dfirst_time=m_pTempoMap->TicksToMsecs(mfxdata.m_lTime)/(double)1000;
     }
     else
     {
      found_last_time=1;
      dlast_time=m_pTempoMap->TicksToMsecs(mfxdata.m_lTime)/(double)1000;
     }
    }
    if(bypass_fts)
    {
     if(fplog)fprintf(fplog," - bypass, sent on unchanged\n"
                       );\
     pqOut->Add(mfxdata);
     continue;
    }
    {\
     DWORD dw=dwData;\
	    WORD lw=LOWORD(dw);
	    BYTE status=LOBYTE(lw);\
     BYTE type=status&0xf0;\
     switch(type)\
      {\
       case MIDI_NOTE_OFF:process=fts_notes;break;\
       case MIDI_NOTE_ON:process=fts_notes;break;\
       case MIDI_SET_PATCH:process=fts_patches;break;\
       case MIDI_PITCH_BEND:process=fts_pitch_bends;break;\
       case MIDI_CONTROLLER:process=fts_controllers;break;\
       case MIDI_AFTERTOUCH:process=fts_aftertouch;break;\
       default:;break;\
      }\
    }
    if(fplog)fprintf(fplog," - %s\n",process?"send on to FTS":"skip"
                      );\
    if(process)
    {
     time_t tstart=clock();;
     static double dlat_max;
     AddFTSData("output",fplog,hwndFTSMFX,pqOut,m_pTempoMap,mfxdata,dwData,WM_DO_MFX_RELAY
               ,&event_processed_in_fts);
     dlat_max=max(dlat_max,(double)(clock()-tstart)/CLOCKS_PER_SEC);
     dlat_max=dlat_max;
    }
    /**
    {
     EVENTS_IN_OUT *pEventsInOut;
     ConstructFileMapping();
     GetFileMapMem(&pEventsInOut);
	    if(pEventsInOut!=NULL)
     {
      pEventsInOut->iversion=FTS_MFX_RELAY_VERSION_1;
      pEventsInOut->iwhat_to_do=MFX_PROCESS_MIDI_KBD;
      pEventsInOut->nOutputVals=0;
      pEventsInOut->dwInput=dwData;
      SendMessage(hwndFTSMFX,WM_DO_MFX_RELAY,0,0);
      event_processed_in_fts=1;
     }
     if(pEventsInOut&&pEventsInOut->nOutputVals>0)
     {
      int jpos=0;
      int nOutputVals=pEventsInOut->nOutputVals;
      DWORD *dwOutput=pEventsInOut->dwOutput;
      for(jpos=0;jpos<nOutputVals;jpos++)
      {
       mfxdata.m_dwData=dwOutput[jpos];
       if(fts_skip_all_controllers)
       {
        DWORD dw=dwOutput[jpos];\
	       WORD lw=LOWORD(dw);
	       BYTE status=LOBYTE(lw);\
        BYTE type=status&0xf0;\
        if(type==MIDI_CONTROLLER)
         continue;
       }
       AddToLog("Output",dwOutput[jpos])
       pqOut->Add(mfxdata);
      }
     }
     if(pEventsInOut)
      FreeFileMapMem(pEventsInOut);       
    }
    **/
			}
		}
  if(fplog)fclose(fplog);
  if(hwndFTSMFX&&event_processed_in_fts&&found_1st_time&&found_last_time)
   SendMessage(hwndFTSMFX,WM_ADD_TO_PLAYING_TIME,(LPARAM)(dlast_time-dfirst_time),0);
	}
 }
 __except(TRUE)
 {
 }
	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// IPersistStream

// WARNING: If you change Save() and Load() and GetSizeMax() to read and write
// different data, you MUST generate a new GUID value for CLSID_FTS_MFX
// and put that in your FTS_MFX.h file.
//
// In other words, this GUID identifies a certain version of the data that your
// plugin saves.

HRESULT CFTS_MFX::Save( IStream* pStm, BOOL bClearDirty )
{
	// Here you should write any data to pStm which you wish to store.
	// Typically you will write values for all of your properties.
	// Save() and Load() are used for the host application to provide "Presets"
	// and to persist your effect in application project files.

 ULONG cbWritten;
	HRESULT const hr = pStm->Write( &prop, sizeof prop, &cbWritten );
	if (SUCCEEDED( hr ))
	{
		if (bClearDirty)
			m_bDirty = FALSE;
		return S_OK;
	}
	else
	{
		// Valid IPersistStream::Write() errors include only STG_E_MEDIUMFULL
		// and STG_E_CANTSAVE. Don't simply return whatever IStream::Write() returned.
		if (STG_E_MEDIUMFULL == hr)
			return STG_E_MEDIUMFULL;
		else
			return STG_E_CANTSAVE;
	}
	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////

HRESULT CFTS_MFX::Load( IStream* pStm )
{
	// Here you should read the data using the format you used in Save().
	// Save() and Load() are used for the host application to provide "Presets"
	// and to persist your effect in application project files.
	
	// Note: IStream::Read() can return S_FALSE, so don't use SUCCEEDED()
	// Valid IPersistStream::Load() errors include only E_OUTOFMEMORY and
	// E_FAIL. Don't simply return whatever IStream::Read() returned.

	ULONG cbRead;
	HRESULT const hr = pStm->Read( &prop, sizeof prop, &cbRead );
	// Note: IStream::Read() can return S_FALSE, so don't use SUCCEEDED()
	if (S_OK == hr && sizeof prop == cbRead)
	{
		m_bDirty = FALSE;
  log_header_up_to_date=0;
  UpdateChecks();
		return S_OK;
	}
	else
	{
		// Valid IPersistStream::Load() errors include only E_OUTOFMEMORY and
		// E_FAIL. Don't simply return whatever IStream::Read() returned.
		return E_FAIL;
	}
	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////

HRESULT CFTS_MFX::GetSizeMax( ULARGE_INTEGER* pcbSize )
{
	// Assign the size of the data you write to the registry in Save() to pcbSize->LowPart
	pcbSize->LowPart = 0;
	pcbSize->HighPart = 0;

	return S_OK;
}

/////////////////////////////////////////////////////////////////////////////
